
#include <cassert>
#include <chrono>
#include <numeric>
#include <thread>
#include <vector>

#include "operator/FC.hpp"

#include "operator/FCImpl.hpp"
#include "operator/FCImpl_kernels.hpp"
#include "utils/Types.h"

Aidge::NbElts_t Aidge::FCImpl_ref_cpp::getNbRequiredData(Aidge::IOIndex_t inputIdx) const
{
    assert(mOp.getInput(inputIdx) && "requires valid input");

    // Requires the whole tensors
    const auto &inputDims
        = std::static_pointer_cast<Tensor>(mOp.getInput(inputIdx))->dims();

    return std::accumulate(
        inputDims.begin(),
        inputDims.end(),
        Aidge::NbElts_t(1),
        std::multiplies<Aidge::NbElts_t>());
}

Aidge::NbElts_t
    Aidge::FCImpl_ref_cpp::getNbRequiredProtected(Aidge::IOIndex_t /*inputIdx*/) const
{
    // for the direct convolution algorithm, convolutions can be in-place, if
    // there is no padding!
    return 0;
}

Aidge::NbElts_t Aidge::FCImpl_ref_cpp::getRequiredMemory(
    IOIndex_t outputIdx, const std::vector<DimSize_t> & /*inputsSize*/) const
{
    // Requires the whole tensors, regardless of available data on inputs
    assert(outputIdx == 0 && "operator has only one output");

    const auto &outputDims = std::static_pointer_cast<Tensor>(mOp.getOutput(0))->dims();
    return std::accumulate(
        outputDims.begin(),
        outputDims.end(),
        static_cast<NbElts_t>(1),
        std::multiplies<NbElts_t>());
}

Aidge::NbElts_t Aidge::FCImpl_ref_cpp::getNbConsumedData(Aidge::IOIndex_t inputIdx) const
{
    assert((inputIdx >= 0) && (static_cast<IONb_t>(inputIdx) < mNbConsumedData.size()));
    return mNbConsumedData[static_cast<std::size_t>(inputIdx)];
}

Aidge::NbElts_t Aidge::FCImpl_ref_cpp::getNbProducedData(Aidge::IOIndex_t outputIdx) const
{
    assert(static_cast<std::size_t>(outputIdx) < mNbProducedData.size());
    return mNbProducedData[static_cast<std::size_t>(outputIdx)];
}

void Aidge::FCImpl_ref_cpp::forward()
{
    // FIXME: uncomment the following code once memory handling will work
    assert(mOp.mInputs[0] && "missing input #0");
    assert(mOp.mInputs[1] && "missing input #1");
    assert(mOp.mInputs[2] && "missing input #2");

    // Find the correct kernel type
    auto kernelFunc = Registrar<FCImplForward_ref_cpp>::create(
        {mOp.mInputs[0]->dataType(),
         mOp.mInputs[1]->dataType(),
         mOp.mInputs[2]->dataType(),
         mOp.mOutput->dataType()});

    // Call kernel
    // if (mOp.mInputs[0]->nbDims() == 4) {
    //     kernelFunc(
    //         mOp.getParams(),
    //         std::static_pointer_cast<Tensor>(mOp.mInputs[0])->dims<4>(),
    //         mOp.mInputs[0]->getImpl()->rawPtr(),
    //         mOp.mInputs[1]->getImpl()->rawPtr(),
    //         mOp.mInputs[2]->getImpl()->rawPtr(),
    //         mOp.mOutput->getImpl()->rawPtr());
    // }
    // else 
    kernelFunc(
        mOp.getParams(),
        mOp.mInputs[0]->dims()[0],
        mOp.mInputs[0]->sizeM1(),
        mOp.mInputs[0]->getImpl()->rawPtr(),
        mOp.mInputs[1]->getImpl()->rawPtr(),
        mOp.mInputs[2]->getImpl()->rawPtr(),
        mOp.mOutput->getImpl()->rawPtr());
    
    

    // FIXME: Dummy wait for some earlier scheduler tests
    std::this_thread::sleep_for(std::chrono::milliseconds(mOp.get<FCParam::OutChannels>()));

    // Update producer-consumer data
    for (IOIndex_t inputIdx = 0; static_cast<std::size_t>(inputIdx) < mNbConsumedData.size(); ++inputIdx)
        mNbConsumedData[inputIdx]
            += getNbRequiredData(static_cast<std::size_t>(inputIdx)); // each input is consumed by the minimum
                                              // amount for a forward pass

    mNbProducedData[0] += getRequiredMemory(0, {});
}

void Aidge::FCImpl_ref_cpp::backward()
{
    printf("Not implemented yet.\n");
}
