/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef __AIDGE_TENSOR_H__
#define __AIDGE_TENSOR_H__

#include <cstring>
#include <set>
#include <memory>
#include <numeric>
#include <string>
#include <vector>

#include "backend/TensorImpl.hpp"
#include "data/Data.hpp"
#include "utils/Registrar.hpp"
#include "utils/Types.h"

namespace Aidge {
// Helper to convert vector to array
template <typename T, typename Iter, std::size_t... Is>
constexpr auto to_array(Iter &iter, std::index_sequence<Is...>) -> std::array<T, sizeof...(Is)> {
    return {{((void)Is, T(*iter++))...}};
}

/**
 * @brief Convert an object with an iterator to an std::array.
 */
template <std::size_t N, typename U = void, typename Iter, typename V = typename std::iterator_traits<Iter>::value_type,
          typename T = std::conditional_t<std::is_same<U, void>{}, V, U>>
constexpr auto to_array(Iter iter) -> std::array<T, N> {
    return to_array<T>(iter, std::make_index_sequence<N>{});
}

namespace detail {

template <class T, std::size_t N, std::size_t... I>
constexpr std::array<std::remove_cv_t<T>, N> to_array_impl(T (&a)[N], std::index_sequence<I...>) {
    return {{a[I]...}};
}

}  // namespace detail

/**
 * @brief Convert a C-stype array into a C++ std::array.
 *
 * @tparam T Data type.
 * @tparam N Number of elements.
 * @param a C-style array to convert.
 * @return constexpr std::array<std::remove_cv_t<T>, N>
 */
template <class T, std::size_t N>
constexpr std::array<std::remove_cv_t<T>, N> to_array(T (&a)[N]) {
    return detail::to_array_impl(a, std::make_index_sequence<N>{});
}

template <typename T, std::size_t N, std::size_t... I>
constexpr std::array<T, N + 1> append(std::array<T, N> a, T t, std::index_sequence<I...>) {
    return std::array<T, N + 1>{a[I]..., t};
}

template <typename T, std::size_t N, std::size_t... I>
constexpr std::array<T, N + 1> append(T t, std::array<T, N> a, std::index_sequence<I...>) {
    return std::array<T, N + 1>{t, a[I]...};
}

/**
 * @brief Create a new array concatenating the initial one with the value to
 * add.
 * @details append([1,2,7], 3) -> [1,2,7,3]
 *
 * @tparam T Data type.
 * @tparam N Number of elements in the initilial array.
 * @param a Initial array.
 * @param t Element to add.
 * @return constexpr std::array<T, N + 1>
 */
template <typename T, std::size_t N>
constexpr std::array<T, N + 1> append(std::array<T, N> a, T t) {
    return append(a, t, std::make_index_sequence<N>());
}

template <typename T, std::size_t N>
constexpr std::array<T, N + 1> append(T t, std::array<T, N> a) {
    return append(t, a, std::make_index_sequence<N>());
}

// Generic helper for initializing a Tensor
template <typename T, std::size_t SIZE_0>
struct Array1D {
    T data[SIZE_0];
};

template <typename T, std::size_t SIZE_0, std::size_t SIZE_1>
struct Array2D {
    T data[SIZE_0][SIZE_1];
};

template <typename T, std::size_t SIZE_0, std::size_t SIZE_1, std::size_t SIZE_2>
struct Array3D {
    T data[SIZE_0][SIZE_1][SIZE_2];
};

template <typename T, std::size_t SIZE_0, std::size_t SIZE_1, std::size_t SIZE_2, std::size_t SIZE_3>
struct Array4D {
    T data[SIZE_0][SIZE_1][SIZE_2][SIZE_3];
};

class Tensor : public Data,
               public Registrable<std::tuple<std::string, DataType>, std::unique_ptr<TensorImpl>(const Tensor &)> {
   private:
    DataType mDataType;
    std::vector<DimSize_t> mDims;
    std::unique_ptr<TensorImpl> mImpl;
    std::shared_ptr<Tensor> mGrad;

    // Cached data
    std::size_t mSize;    // number of elements in the tensor
    std::size_t mSizeM1;  // for a tensor of N dimensions, number of elements in the N-1
                     // first dimensions

   public:
    static constexpr const char *Type = "Tensor";

    Tensor(DataType dataType = DataType::Float32) : Data(Type), mDataType(dataType), mDims({}), mSize(0), mSizeM1(0) {
        // ctor
    }

    template <typename T, std::size_t SIZE_0>
    constexpr Tensor(Array1D<T, SIZE_0> &&arr)
        : Data(Type),
          mDims({SIZE_0}),
          mDataType(NativeType<T>::type),
          mImpl(Registrar<Tensor>::create({"cpu", NativeType<T>::type})(*this)),
          mSize(SIZE_0),
          mSizeM1(SIZE_0) {
        mImpl->copy(&arr.data[0], SIZE_0);
    }

    template <typename T, std::size_t SIZE_0>
    constexpr Tensor &operator=(Array1D<T, SIZE_0> &&arr) {
        resize({SIZE_0});
        if (!mImpl) {
            mImpl = Registrar<Tensor>::create({"cpu", NativeType<T>::type})(*this);
        }
        mImpl->copy(&arr.data[0], SIZE_0);
        return *this;
    }
    template <typename T, std::size_t SIZE_0, std::size_t SIZE_1>
    constexpr Tensor(Array2D<T, SIZE_0, SIZE_1> &&arr)
        : Data(Type),
          mDims({SIZE_0, SIZE_1}),
          mDataType(NativeType<T>::type),
          mImpl(Registrar<Tensor>::create({"cpu", NativeType<T>::type})(*this)),
          mSize(SIZE_0 * SIZE_1),
          mSizeM1(SIZE_1) {
        mImpl->copy(&arr.data[0][0], SIZE_0 * SIZE_1);
    }

    template <typename T, std::size_t SIZE_0, std::size_t SIZE_1>
    constexpr Tensor &operator=(Array2D<T, SIZE_0, SIZE_1> &&arr) {
        resize({SIZE_0, SIZE_1});
        if (!mImpl) {
            mImpl = Registrar<Tensor>::create({"cpu", NativeType<T>::type})(*this);
        }
        mImpl->copy(&arr.data[0][0], SIZE_0 * SIZE_1);
        return *this;
    }
    template <typename T, std::size_t SIZE_0, std::size_t SIZE_1, std::size_t SIZE_2>
    constexpr Tensor(Array3D<T, SIZE_0, SIZE_1, SIZE_2> &&arr)
        : Data(Type),
          mDims({SIZE_0, SIZE_1, SIZE_2}),
          mDataType(NativeType<T>::type),
          mImpl(Registrar<Tensor>::create({"cpu", NativeType<T>::type})(*this)),
          mSize(SIZE_0 * SIZE_1 * SIZE_2),
          mSizeM1(SIZE_1 * SIZE_2) {
        mImpl->copy(&arr.data[0][0][0], SIZE_0 * SIZE_1 * SIZE_2);
    }

    template <typename T, std::size_t SIZE_0, std::size_t SIZE_1, std::size_t SIZE_2>
    constexpr Tensor &operator=(Array3D<T, SIZE_0, SIZE_1, SIZE_2> &&arr) {
        resize({SIZE_0, SIZE_1, SIZE_2});
        if (!mImpl) {
            mImpl = Registrar<Tensor>::create({"cpu", NativeType<T>::type})(*this);
        }
        mImpl->copy(&arr.data[0][0][0], SIZE_0 * SIZE_1 * SIZE_2);
        return *this;
    }
    template <typename T, std::size_t SIZE_0, std::size_t SIZE_1, std::size_t SIZE_2, std::size_t SIZE_3>
    constexpr Tensor(Array4D<T, SIZE_0, SIZE_1, SIZE_2, SIZE_3> &&arr)
        : Data(Type),
          mDims({SIZE_0, SIZE_1, SIZE_2, SIZE_3}),
          mDataType(NativeType<T>::type),
          mImpl(Registrar<Tensor>::create({"cpu", NativeType<T>::type})(*this)),
          mSize(SIZE_0 * SIZE_1 * SIZE_2 * SIZE_3),
          mSizeM1(SIZE_1 * SIZE_2 * SIZE_3) {
        mImpl->copy(&arr.data[0][0][0][0], SIZE_0 * SIZE_1 * SIZE_2 * SIZE_3);
    }

    template <typename T, std::size_t SIZE_0, std::size_t SIZE_1, std::size_t SIZE_2, std::size_t SIZE_3>
    constexpr Tensor &operator=(Array4D<T, SIZE_0, SIZE_1, SIZE_2, SIZE_3> &&arr) {
        resize({SIZE_0, SIZE_1, SIZE_2, SIZE_3});
        if (!mImpl) {
            mImpl = Registrar<Tensor>::create({"cpu", NativeType<T>::type})(*this);
        }
        mImpl->copy(&arr.data[0][0][0][0], SIZE_0 * SIZE_1 * SIZE_2 * SIZE_3);
        return *this;
    }

    bool operator==(const Tensor &otherTensor) const {
        if ((!mImpl && !otherTensor.mImpl) || (dataType() != otherTensor.dataType()) ||
            (dims() != otherTensor.dims()) || (mImpl->backend() != otherTensor.mImpl->backend())) {
            return false;
        }
        return *mImpl == *(otherTensor.mImpl);
    }

    inline void setBackend(const std::string &name) {
        if (mImpl) {
            if (strcmp(mImpl->backend(), name.c_str()) != 0) {
                // Backend change: create new impl, copy from old to new and replace
                // impl
                std::unique_ptr<TensorImpl> newImpl = Registrar<Tensor>::create({name, mDataType})(*this);
                newImpl->copy(mImpl->rawPtr(), size());
                mImpl = std::move(newImpl);
            }
        } else
            mImpl = Registrar<Tensor>::create({name, mDataType})(*this);
    }
    static std::set<std::string> getAvailableBackends(){
        std::set<std::string> backendsList;
        for(std::tuple<std::string, DataType> tupleKey : Registrar<Tensor>::getKeys())
            backendsList.insert(std::get<0>(tupleKey));
        return backendsList;
    }

    constexpr DataType dataType() const { return mDataType; }

    /**
     * @brief Set the DataType of the Tensor and converts data
     * if the Tensor has already been initialized.
     * @param dt DataType.
     */
    void setDatatype(const DataType dt) {
        if (mImpl && (dataType() != dt)) {
            // get ptr before changing Tensor backend or the type difference will trigger a warning
            const void *data = mImpl->rawPtr();
            mDataType = dt;
            std::unique_ptr<TensorImpl> newImpl = Registrar<Tensor>::create({mImpl->backend(), dt})(*this);
            newImpl->copy(data, size());  // /!\ it does not cast data but reinterpret them
            mImpl = std::move(newImpl);
        }
        mDataType = dt;
    }

    constexpr const std::unique_ptr<TensorImpl> &getImpl() { return mImpl; }

    bool hasImpl() 
    {
        return (mImpl) ? true : false;
    }

    inline std::size_t nbDims() const { return mDims.size(); }

    template <DimIdx_t DIM>
    constexpr std::array<DimSize_t, DIM> dims() const {
        assert(DIM == mDims.size() && "wrong number of dimensions");
        return to_array<DIM>(mDims.cbegin());
    }

    constexpr const std::vector<DimSize_t> &dims() const { return mDims; }

    constexpr std::size_t size() const { return mSize; }

    constexpr std::size_t sizeM1() const { return mSizeM1; }

// deducing std::array size_type and declaring DIM accordingly
    template <std::array<DimSize_t, 1>::size_type DIM>
    void resize(const std::array<DimSize_t, DIM> &dims) {
        static_assert(DIM<=MaxDim,"Too many tensor dimensions required by resize, not supported");
        mDims.assign(dims.begin(), dims.end());
        computeSize();
    }
    void resize(const std::vector<DimSize_t> &dims) {
        mDims = dims;
        computeSize();
    }
    bool empty() const { return mDims.empty(); }
    template <typename expectedType, std::array<std::size_t, 1>::size_type DIM>
    constexpr expectedType &get(std::array<std::size_t, DIM> idx) {
        assert(DIM == mDims.size());
        assert(mImpl);
        std::size_t unfoldedIdx = 0;
        for (std::size_t i = 0; i < DIM - std::size_t(1); ++i) {
            unfoldedIdx = (unfoldedIdx + idx[i]) * mDims[i + 1];
        }
        unfoldedIdx += idx[DIM - 1];
        return static_cast<expectedType *>(mImpl->rawPtr())[unfoldedIdx];
    }

    std::string toString() {
        std::string res;
        std::size_t dim = 0;
        std::size_t *dimVals = new std::size_t[nbDims()];
        for (std::size_t i = 0; i < nbDims(); ++i) {
            dimVals[i] = 0;
        }
        std::size_t counter = 0;
        res += "{\n";
        if (nbDims()>=2){
            while (counter < mSize) {
                std::string spaceString = std::string((dim+1)<<1,' ');
                if (dim < nbDims()-2) {
                    if (dimVals[dim] == 0) {
                        res += spaceString + "{\n";
                        ++dim;
                    } else if (dimVals[dim] < static_cast<std::size_t>(dims()[dim])) {
                        res += spaceString + "},\n" + spaceString + "{\n";
                        ++dim;
                    } else {
                        res += spaceString + "}\n";
                        dimVals[dim--] = 0;
                        dimVals[dim]++;
                    }
                } else {
                    for (; dimVals[dim] < static_cast<std::size_t>(dims()[dim]); ++dimVals[dim]) {
                        res += spaceString + "{";
                        for (DimSize_t j = 0; j < dims()[dim + 1] - 1; ++j) {
                            switch (mDataType)
                            {
                            case DataType::Int32:
                                res += " " + std::to_string(static_cast<int *>(mImpl->rawPtr())[counter++]) + ",";
                                break;
                            case DataType::Float64:
                                res += " " + std::to_string(static_cast<double *>(mImpl->rawPtr())[counter++]) + ",";
                                break;
                            default:
                                res += " " + std::to_string(static_cast<float *>(mImpl->rawPtr())[counter++]) + ",";
                                break;
                            }
                        }
                        switch (mDataType)
                        {
                        case DataType::Int32:
                            res += " " + std::to_string(static_cast<int *>(mImpl->rawPtr())[counter++]) + "}";
                            break;
                        case DataType::Float64:
                            res += " " + std::to_string(static_cast<double *>(mImpl->rawPtr())[counter++]) + "}";
                            break;
                        default:
                            res += " " + std::to_string(static_cast<float *>(mImpl->rawPtr())[counter++]) + "}";
                            break;
                        }
                        if (dimVals[dim] < static_cast<std::size_t>(dims()[dim] - 1)) {
                            res += ",";
                        }
                        res += "\n";
                    }
                    dimVals[dim--] = 0;
                    dimVals[dim]++;
                }
            }
            for(int i = static_cast<int>(dim); i>=0; --i) {
                res += std::string((dim+1)<<1,' ') + "}\n";
            }
        }else{
            for (DimSize_t j = 0; j < dims()[0]; ++j) {
                switch (mDataType)
                {
                case DataType::Int32:
                    res += " " + std::to_string(static_cast<int *>(mImpl->rawPtr())[j]) + ((j < dims()[0]-1) ? "," : "\n");
                    break;
                case DataType::Float64:
                    res += " " + std::to_string(static_cast<double *>(mImpl->rawPtr())[j]) + ((j < dims()[0]-1) ? "," : "\n");
                    break;
                default:
                    res += " " + std::to_string(static_cast<float *>(mImpl->rawPtr())[j]) + ((j < dims()[0]-1) ? "," : "\n");
                    break;
                }
            }
        }
        
        
        res += "}";
        return res;
    }

    inline void print() { printf("%s\n", toString().c_str()); }

    std::shared_ptr<Tensor> grad() {
        if (!mGrad) {
            mGrad = std::make_shared<Tensor>(mDataType);
            mGrad->resize(mDims);

            if (mImpl) mGrad->setBackend(mImpl->backend());
        }

        return mGrad;
    }

private:
    ///\bug not protected against overflow, see ThaliaCommonPack for a solution
    std::size_t computeSize() {
        if (mDims.empty()) {
            mSizeM1 = DimSize_t(0);
            mSize = DimSize_t(0);
        }
        else if (mDims.size() == 1)
        {
            mSizeM1 = mDims[0];
            mSize = mDims[0];
        }
        else {
            mSizeM1 = std::accumulate(++mDims.begin(),mDims.end(), DimSize_t(1), std::multiplies<DimSize_t>());
            mSize = static_cast<std::size_t>(mSizeM1 * mDims[0]);
        }
        
        return mSize;
    }
};

    // template <class T>
    // class Tensor
    // {
    // private:
    //     vector<T> mData;
    //     vector<short> mDims;
    //     short mSize;

    //     /// @brief Size of the n-1 last dimensions
    //     /// @return For a Tensor T of dimension [N,C,H,W]
    //     /// returns C*H*W
    //     short sizeM1();

    // public:
    //     Tensor();
    //     Tensor(const Tensor<T>& other);
    //     Tensor(std::initializer_list<short> dims,
    //            const T &value = T());
    //     Tensor(std::vector<short> dims,
    //             const T& value = T());
    //     ~Tensor() = default;

    //     /// @brief Fills the Tensor with one single value
    //     /// @param value new value
    //     void fill(const T& value);
    //     /// @brief Number of elements in the Tensor
    //     short size();
    // };

    // template <class T>
    // Tensor<T>::Tensor()
    //     : mData(std::make_shared<std::vector<T>>(std::vector<T>())),
    //       mDims({}),
    //       mSize(0)
    // {
    //     // ctor
    // }

    // template <class T>
    // Tensor<T>::Tensor(std::initializer_list<short> dims,
    //                     const T& value)
    // {
    //     mDims = dims;
    //     mSize = (!dims.empty()) ? std::accumulate(dims.begin(),
    //                                                dims.end(),
    //                                                1U,
    //                                                std::multiplies<short>()) :
    //                                                0U;
    //     mData = std::make_shared<std::vector<T>>(std::vector<T>(mSize, value));
    // }

    // template <class T>
    // Tensor<T>::Tensor(std::vector<short> dims,
    //                     const T& value)
    // {
    //     mDims = dims;
    //     mSize = (!dims.empty()) ? std::accumulate(dims.begin(),
    //                                                dims.end(),
    //                                                1U,
    //                                                std::multiplies<short>()) :
    //                                                0U;
    //     mData = std::make_shared<std::vector<T>>(std::vector<T>(mSize, value));
    // }

    // template <class T>
    // Tensor<T>::Tensor(const Tensor<T>& other)
    //     : mDims(other.mDims),
    //       mData(other.mData),
    //       mSize(other.size())
    // {
    //     // ctor by copy
    //     // Warning: a tensor "B" created from another tensor "A" with
    //     // this copy-ctor will share the same mData as "A".
    //     // Indeed, this implementation only copies the address of mData
    //     // because mData is a shared pointer. Thus if one of the values of B
    //     // is changed, then the value of A is also changed.
    //     // If you want to create a tensor from another tensor with different
    //     // memory locations, please use Tensor::clone()
    // }

    // template <class T>
    // short Tensor<T>::size() {
    //     return mSize;
    // }

    // template <class T>
    // short Tensor<T>::sizeM1() {
    //     return (!mDims.empty()) ? std::accumulate(++mDims.begin(),
    //                                                mDims.end(),
    //                                                1U,
    //                                                std::multiplies<short>()) :
    //                                                0U;
    // }

    // template <class T>
    // void Tensor<T>::fill(const T& value) {
    //     std::fill(mData.begin(), mData.end(), value);
    // }
}  // namespace Aidge

#endif /* __AIDGE_TENSOR_H__ */
