/**
 * @file
 * @brief 
 * @version file 1.0.0
 * @author vl241552
 * @copyright
 *  Copyright (c) 2023 CEA, LIST, Embedded Artificial Intelligence Laboratory. All
 *  rights reserved.
 */



#ifndef _AIDGE_CONDITIONAL_LEXER_H_
#define _AIDGE_CONDITIONAL_LEXER_H_

#include <string>
#include <regex>
#include <memory> // for shared_ptr


#include <stdexcept> //error
#include <sstream>

namespace Aidge{



/**
 * @brief enum for all types of token use in the parsing 
 * 7-5 type
 * 4-0 id
*/
enum class ConditionalTokenTypes
{
    NOT                 = (1 << 4)+3, /**< ! */ 
    AND                 = (1 << 4)+2, /**< && */ 
    OR                  = (1 << 4)+1, /**< || */ 

    EQ                  = (1 << 5)+1, /**< == */ 
    NEQ                 = (1 << 5)+2, /**< != */ 

    KEY                 = (1 << 6) +1 , /**< [A-Za-z][A-Za-z0-9_]* */
    INTEGER             = (1 << 6) +2 , /**< [0-9]+ */
    FLOAT               = (1 << 6) +3 , /**< [0-9]+\.[0-9]* */
    STRING              = (1 << 6) +4 , /**< \'.*\' */
    BOOL                = (1 << 6) +5 , /**< true|false */
    NODE                = (1 << 6) +6 , /**< \$ */
    LAMBDA              = (1 << 6) +7 , /**< [A-Za-z][A-Za-z0-9_]*\( */

    ARGSEP              = (1<<7) +1, /**< , */
    LPAREN              = (1<<7) +2, /**< \( */
    RPAREN              = (1<<7) +3, /**< \) */
    STOP                = 0,
};



/**
 * @brief token holder
*/
class ConditionalToken
{
    public:
    /**
     * @brief Token container
     * @param type one of the token type
     * @param lexeme String representing aditional information of the token
     */
    ConditionalToken(const ConditionalTokenTypes type , const std::string lexeme );
    /**
     * @brief get the lexeme
     * @return std::string 
     */
    const std::string getLexeme(void);

    /**
     * @brief get the token type
     * 
     * @return ConditionalTokenTypes 
     */
    const ConditionalTokenTypes getType(void);

    /**
     * @brief copy the token
     * @return deep copy of the token
     */
    std::shared_ptr<Aidge::ConditionalToken> copy();


    std::ostringstream rep(void);
    private:

    /**
     * @brief additional information of the token
     */
    const std::string mLexeme;

    /**
     * @brief type of the token
     * @see ConditionalTokenTypes
     */
    const ConditionalTokenTypes mType;

};



class ConditionalLexer
{

public:
ConditionalLexer( const std::string ConditionalExpressions );

/**
 * @brief Get the next token on the ConditionalExpressions
 * @return ConditionalToken 
 */
std::shared_ptr<ConditionalToken> getNextToken(void);
/**
 * @brief Restart at the start of the ConditionalExpressions 
 * 
 */
void rstPosition(void);

/**
 * @brief Test if the string is completely read 
 * @return bool 
 */
bool isEnd(void);


/**
 * @brief Get the representation of the class
 * @return string 
 */
const std::string rep(){
   return mConditionalExpressions;
}

private:

/**
 * @brief Constructs an error message to display the character not understood by the lexer 
 * @return error mesage 
 */
std::runtime_error badTokenError(const std::string& currentChars,std::size_t position);

/**
 * @brief The expression of the test to be performed on the nodes
 */
const std::string mConditionalExpressions;
/**
 * @brief The lexer's current position in mConditionalExpressions
 */
std::size_t mPosition;

};

/////////////////////////////////////


}

#endif //_AIDGE_CONDITIONAL_LEXER_H_