/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef __AIDGE_PARAMETER_H__
#define __AIDGE_PARAMETER_H__

#include <tuple>
#include <cassert>
#include <cstddef>

namespace {
// This is the type that will hold all the strings. Each enumerate type will
// declare its own specialization.
template <typename T> struct EnumStrings {
    static const char* const data[];
};
}

namespace Aidge {
template<class T, std::size_t N>
constexpr std::size_t size(T (&)[N]) { return N; }

template <class PARAM_ENUM, class ...T>
class Parameterizable {
public:
    typedef std::tuple<T...> Parameters;

    // Helper class to pass to the constructor
    template <PARAM_ENUM paramEnum>
    class param {
    public:
        constexpr param(const typename std::tuple_element<static_cast<std::size_t>(paramEnum),std::tuple<T...>>::type& v) : value(v) {}
        const typename std::tuple_element<static_cast<std::size_t>(paramEnum),std::tuple<T...>>::type value;
    };

/*
    // Direct tuple initialization
    Parameterizable(T... params) : mParams({params...}) {

    }
*/

    // Constructor for parameters initialization.
    // Compile-time garantee that every parameter is initialized.
    template <PARAM_ENUM ...paramEnum> // non-type parameter pack
    constexpr Parameterizable(const param<paramEnum>&&... params) {
        // Check number of params consistency
        static_assert(sizeof...(params) == std::tuple_size<std::tuple<T...>>::value, "wrong number of parameters in constructor");
        // static_assert(size(EnumStrings<PARAM_ENUM>::data) == std::tuple_size<std::tuple<T...>>::value, "wrong number of parameters in enum string");

        // Check no duplicates
        constexpr std::array<PARAM_ENUM, std::tuple_size<std::tuple<T...>>::value> pe = { paramEnum... };
        static_assert(!hasDuplicates(pe), "duplicate parameter"); // requires C++14

        // Init params with constructor arguments
        const std::array<PARAM_ENUM, std::tuple_size<std::tuple<T...>>::value> p = { ((void)(get<paramEnum>() = params.value), paramEnum) ... };
        (void)p; // avoid unused warning
    }

    // Compile-time access with enum
    template <PARAM_ENUM paramEnum>
    constexpr typename std::tuple_element<static_cast<std::size_t>(paramEnum),std::tuple<T...>>::type& get() {
        return std::get<static_cast<std::size_t>(paramEnum)>(mParams);
    }
    
    template <PARAM_ENUM paramEnum>
    constexpr const typename std::tuple_element<static_cast<std::size_t>(paramEnum),std::tuple<T...>>::type& get() const {
        return std::get<static_cast<std::size_t>(paramEnum)>(mParams);
    }

    // Runtime access with enum
    template <typename R>
    constexpr R& get(PARAM_ENUM paramEnum) {
        return get<R>(static_cast<std::size_t>(paramEnum));
    }

    template <typename R>
    constexpr const R& get(PARAM_ENUM paramEnum) const {
        return get<R>(static_cast<std::size_t>(paramEnum));
    }

    // Runtime existance check with name
    constexpr bool isParam(const char* name) const {
        for (std::size_t i = 0; i < size(EnumStrings<PARAM_ENUM>::data); ++i) {
            if (strcmp(EnumStrings<PARAM_ENUM>::data[i], name) == 0) {
                return true;
            }
        }

        return false;
    }

    // Runtime access with name
    template <typename R>
    constexpr R& get(const char* name) {
        for (std::size_t i = 0; i < size(EnumStrings<PARAM_ENUM>::data); ++i) {
            if (strcmp(EnumStrings<PARAM_ENUM>::data[i], name) == 0) {
                return get<R>(i);
            }
        }

        assert(false && "parameter not found");
    }

    template <typename R, std::size_t SIZE = std::tuple_size<std::tuple<T...>>::value-1>
    constexpr typename std::enable_if<(SIZE > 0), R&>::type get(std::size_t i) {
        if (i == SIZE) {
            if (std::is_same<R, typename std::tuple_element<SIZE,std::tuple<T...>>::type>::value) {
                return reinterpret_cast<R&>(std::get<SIZE>(mParams));
            }
            else {
                assert(false && "wrong parameter type");
            }
        }
        else {
            return get<R, SIZE-1>(i);
        }
    }

    template <typename R, std::size_t SIZE = std::tuple_size<std::tuple<T...>>::value-1>
    constexpr typename std::enable_if<(SIZE <= 0), R&>::type get(std::size_t i) {
        assert(false && "parameter not found");
    }

    constexpr const std::tuple<T...>& getParams() const {
        return mParams;
    }

private:
    template <typename V, std::size_t N>
    static constexpr bool hasDuplicates(const std::array<V, N>& array) {
        for (std::size_t i = 1; i < N; i++) {
            for (std::size_t j = 0; j < i; j++) {
                if (array[i] == array[j]) {
                    return true;
                }
            }
        }

        return false;
    }

    std::tuple<T...> mParams;
};
}

#endif /* __AIDGE_PARAMETER_H__ */
