/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#include <pybind11/pybind11.h>
#include <pybind11/stl.h>
#include <pybind11/operators.h>
#include <pybind11/numpy.h>

#include "data/Tensor.hpp"
#include "data/Data.hpp"
#include "utils/Registrar.hpp"
#include "utils/Types.h"
#include "backend/TensorImpl.hpp"

namespace py = pybind11;
namespace Aidge {


template<typename T>
void addCtor(py::class_<Tensor,
                        std::shared_ptr<Tensor>, 
                        Data, 
                        Registrable<std::tuple<std::string, DataType>, 
                        std::unique_ptr<TensorImpl>(const Tensor&)>>& mTensor){
    mTensor.def(py::init([]( py::array_t<T, py::array::c_style | py::array::forcecast> b) {
        /* Request a buffer descriptor from Python */
        py::buffer_info info = b.request();
        Tensor* newTensor = new Tensor();
        newTensor->setDatatype(NativeType<T>::type);
        const std::vector<DimSize_t> dims(info.shape.begin(), info.shape.end());
        newTensor->resize(dims);
        // TODO : Find a better way to choose backend
        std::set<std::string> availableBackends = Tensor::getAvailableBackends();
        if (availableBackends.find("cpu") != availableBackends.end()){
            newTensor->setBackend("cpu");
            newTensor->getImpl()->setRawPtr(static_cast<T*>(info.ptr));
        }else{
            printf("Warning : Could not use aidge_ref_cpp backend, verify you have `import aidge_ref_cpp`\n");
        }
        
        return newTensor;
    }));
}


void init_Tensor(py::module& m){
    py::class_<Registrable<std::tuple<std::string, DataType>, 
    std::unique_ptr<TensorImpl>(const Tensor&)>,
    std::shared_ptr<Registrable<std::tuple<std::string, DataType>, 
    std::unique_ptr<TensorImpl>(const Tensor&)>>>(m,"TensorRegistrable");

    py::class_<Tensor, std::shared_ptr<Tensor>, 
               Data, 
               Registrable<std::tuple<std::string, DataType>, 
               std::unique_ptr<TensorImpl>(const Tensor&)>> pyClassTensor
        (m,"Tensor", py::multiple_inheritance(), py::buffer_protocol());

    pyClassTensor.def(py::init<>())
    .def("set_backend", &Tensor::setBackend, py::arg("name"))
    .def("dims", (const std::vector<DimSize_t>& (Tensor::*)()const) &Tensor::dims)
    .def("dtype", &Tensor::dataType)
    .def("size", &Tensor::size)
    .def("resize", (void (Tensor::*)(const std::vector<DimSize_t>&)) &Tensor::resize)
    .def("has_impl", &Tensor::hasImpl)
    .def_static("get_available_backends", &Tensor::getAvailableBackends)
    .def("__str__", [](Tensor& b) {
        return b.toString();
    })
    .def("__len__", [](Tensor& b) -> size_t{
        return b.size();
    })
    .def("__getitem__", [](Tensor& b, size_t idx)-> py::object {
        // TODO : Should return error if backend not compatible with get
        if (idx >= b.size()) throw py::index_error();
        switch(b.dataType()){
            case DataType::Float32:
                return py::cast(static_cast<float*>(b.getImpl()->rawPtr())[idx]);
            case DataType::Int32:
                return py::cast(static_cast<int*>(b.getImpl()->rawPtr())[idx]);
            default:
                return py::none();
        }
    })
    .def_buffer([](Tensor& b) -> py::buffer_info {
        const std::unique_ptr<TensorImpl>& tensorImpl = b.getImpl();

        std::vector<ssize_t> dims;
        std::vector<ssize_t> strides;
        ssize_t stride = tensorImpl->scalarSize();

        for (unsigned int dim = 0; dim < b.nbDims(); ++dim) {
            dims.push_back(b.dims()[dim]);
            strides.push_back(stride);
            stride *= b.dims()[dim];
        }
        std::reverse(dims.begin(), dims.end());
        std::reverse(strides.begin(), strides.end());

        std::string dataFormatDescriptor;
        switch(b.dataType()){
            case DataType::Float32:
                dataFormatDescriptor = py::format_descriptor<float>::format();
                break;
            case DataType::Int32:
                dataFormatDescriptor = py::format_descriptor<int>::format();
                break;
            default:
                throw py::value_error("Unsupported data format");
        }

        return py::buffer_info(
            tensorImpl->rawPtr(),                       /* Pointer to buffer */
            tensorImpl->scalarSize(),                   /* Size of one scalar */
            dataFormatDescriptor,                /* Python struct-style format descriptor */
            b.nbDims(),                                 /* Number of dimensions */
            dims,                                       /* Buffer dimensions */
            strides                                     /* Strides (in bytes) for each index */
        );
    });

    // TODO : If the ctor with the right data type does not exist, pybind will always convert the data to INT !
    // Need to find a way to avoid this !
    addCtor<int>(pyClassTensor);
    addCtor<float>(pyClassTensor);
// #if SIZE_MAX != 0xFFFFFFFF
    addCtor<double>(pyClassTensor);
// #endif
    
}
}
