#include <iostream>
#include <map>
#include <memory>
#include <vector>
#include <utility>
#include <cassert>

#include <catch2/catch_test_macros.hpp>
//test
#include "graphmatching/GRegex.hpp"
#include "graphmatching/StmFactory.hpp"
#include "graphmatching/SeqStm.hpp"
#include "graphmatching/NodeRegex.hpp"
#include "graphmatching/Match.hpp"
//use
#include "backend/OperatorImpl.hpp"
#include "operator/GenericOperator.hpp"
#include "operator/Producer.hpp"
#include "graph/GraphView.hpp"

using namespace Aidge;

TEST_CASE("Create good init GRegex", "[GRegex]") {
    // init all input for GRegex
    // Nodes Regex map : std::map<std::string,NodeRegex*>& nodesRegex
    // Sequential Regex vector : std::vector<std::string>& seqRegexps

    // init the Nodes Regex map 
    std::map<std::string,NodeRegex*> nodesRegex ;
    std::vector<std::string> nodeTypeKey {"A","B","C"};
    for (const std::string& key : nodeTypeKey) {
        nodesRegex[key] = new NodeRegex(key);
    }

    // init the Sequential Regex vector 
    std::vector<std::string> seqRegex;
    seqRegex.push_back("A->B;");

    // Instanciate a GRegex
    GRegex GReg(nodesRegex, seqRegex);

    // Perform tests
    REQUIRE(GReg.getStmInit().size() == 1);
    REQUIRE(GReg.getStmFab().getNumberOfStm() == 1);
}


TEST_CASE("Function matchFromStartNodes | One Match of Nodes sequence", "[GRegex]") {
    // init all input for GRegex
    // Nodes Regex map : std::map<std::string,NodeRegex*>& nodesRegex
    // Sequential Regex vector : std::vector<std::string>& seqRegexps

    // init the Nodes Regex map 
    std::map<std::string,NodeRegex*> nodesRegex ;
    std::vector<std::string> nodeTypeKey {"Conv","BN","ReLU"};
    for (const std::string& key : nodeTypeKey) {
        nodesRegex[key] = new NodeRegex(key); 
    }
    // init the Sequential Regex vector 
    std::vector<std::string> seqRegex;
    seqRegex.push_back("Conv->BN->ReLU;");

    // Instanciate a GRegex
    GRegex GReg(nodesRegex, seqRegex);

    std::shared_ptr<GraphView> g1 = std::make_shared<GraphView>("TestGraph");
    std::shared_ptr<Node> Conv1 = GenericOperator("Conv", 1, 1, 1);
    std::shared_ptr<Node> BN1 = GenericOperator("BN", 1, 1, 1);
    std::shared_ptr<Node> ReLU1 = GenericOperator("ReLU", 1, 1, 1);
    std::shared_ptr<Node> Random = GenericOperator("Random", 1, 1, 1);
    std::shared_ptr<Node> Random2 = GenericOperator("Random2", 1, 1, 1);


    g1->add(Conv1);
    g1->addChild(BN1, Conv1);
    g1->addChild(ReLU1, BN1);
    g1->addChild(Random, ReLU1);
    //g1->addChild(BN1, Random2);

    std::vector<std::shared_ptr<Node>> startNodes1;
    std::set<std::shared_ptr<Node>> result;

    startNodes1.push_back(Conv1);
    result = GReg.matchFromStartNodes(startNodes1, g1);

    std::set<std::shared_ptr<Node>> true_result;
    true_result.insert(Conv1);
    true_result.insert(BN1);
    true_result.insert(ReLU1);

    // Perform tests
    REQUIRE(result == true_result);
}

TEST_CASE("Function matchFromStartNodes | One Match of parallel branches ", "[GRegex]") {
    // init all input for GRegex
    // Nodes Regex map : std::map<std::string,NodeRegex*>& nodesRegex
    // Sequential Regex vector : std::vector<std::string>& seqRegexps

    // init the Nodes Regex map 
    std::map<std::string,NodeRegex*> nodesRegex ;
    std::vector<std::string> nodeTypeKey {"Add","FC","Conv"};
    for (const std::string& key : nodeTypeKey) {
        nodesRegex[key] = new NodeRegex(key); 
    }

    // init the Sequential Regex vector 
    std::vector<std::string> seqRegex;
    seqRegex.push_back("Add#->Conv;");
    seqRegex.push_back("Add#->FC;");

    // Instanciate a GRegex
    GRegex GReg(nodesRegex, seqRegex);

    // Instanciate a graphView
    std::shared_ptr<GraphView> g1 = std::make_shared<GraphView>("TestGraph");
    std::shared_ptr<Node> Random0 = GenericOperator("Random", 1, 1, 1);
    std::shared_ptr<Node> Add1 = GenericOperator("Add", 1, 1, 1);
    std::shared_ptr<Node> Conv1 = GenericOperator("Conv", 1, 1, 1);
    std::shared_ptr<Node> BN1 = GenericOperator("BN", 1, 1, 1);
    std::shared_ptr<Node> ReLU1 = GenericOperator("ReLU", 1, 1, 1);
    std::shared_ptr<Node> FC1 = GenericOperator("FC", 1, 1, 1);
    std::shared_ptr<Node> Random = GenericOperator("Random", 1, 1, 1);

    g1->add(Random0);
    g1->addChild(Add1, Random0);
    g1->addChild(Conv1, Add1);
    g1->addChild(BN1, Conv1);
    g1->addChild(ReLU1, BN1);
    g1->addChild(FC1, Add1);
    g1->addChild(Random, FC1);

    // Test 1 : Find the match
    std::vector<std::shared_ptr<Node>> startNodes;
    std::set<std::shared_ptr<Node>> result;

    startNodes.push_back(Add1);
    startNodes.push_back(Add1);
    result = GReg.matchFromStartNodes(startNodes, g1);

    std::set<std::shared_ptr<Node>> true_result;
    true_result.insert(Add1);
    true_result.insert(Conv1);
    true_result.insert(FC1);

    // Test 2 : Return an empty set when the start nodes are wrong
    std::vector<std::shared_ptr<Node>> wrong_startNodes;
    std::set<std::shared_ptr<Node>> wrong_start_result;
    std::set<std::shared_ptr<Node>> empty_result;

    wrong_startNodes.push_back(Random0);
    wrong_startNodes.push_back(Random0);
    wrong_start_result = GReg.matchFromStartNodes(wrong_startNodes, g1);

    // Perform tests
    REQUIRE(result == true_result);
    REQUIRE(wrong_start_result == empty_result);
}

/*
TEST_CASE("Function matchFromStartNodes | Match a sequence with quantifier ", "[GRegex]") {
    // init all input for GRegex
    // Nodes Regex map : std::map<std::string,NodeRegex*>& nodesRegex
    // Sequential Regex vector : std::vector<std::string>& seqRegexps

    // init the Nodes Regex map 
    std::map<std::string,NodeRegex*> nodesRegex ;
    std::vector<std::string> nodeTypeKey {"FC"};
    for (const std::string& key : nodeTypeKey) {
        nodesRegex[key] = new NodeRegex(key); 
    }
    
    // init the Sequential Regex vector 
    std::vector<std::string> seqRegex;
    seqRegex.push_back("FC+;");

    // Instanciate a GRegex
    GRegex GReg(nodesRegex, seqRegex);

    
    // Instanciate a graphView
    std::shared_ptr<GraphView> g1 = std::make_shared<GraphView>("TestGraph");
    std::shared_ptr<Node> Random0 = GenericOperator("Random", 1, 1, 1);
    std::shared_ptr<Node> FC1 = GenericOperator("FC", 1, 1, 1);
    std::shared_ptr<Node> FC2 = GenericOperator("FC", 1, 1, 1);
    std::shared_ptr<Node> FC3 = GenericOperator("FC", 1, 1, 1);
    std::shared_ptr<Node> ReLU1 = GenericOperator("ReLU", 1, 1, 1);

    g1->add(Random0);
    g1->addChild(FC1, Random0);
    g1->addChild(FC2, FC1);
    g1->addChild(FC3, FC2);
    g1->addChild(ReLU1, FC3);

    // Test 1 : Find the match
    std::vector<std::shared_ptr<Node>> startNodes;
    std::set<std::shared_ptr<Node>> result;

    startNodes.push_back(FC1);
    result = GReg.matchFromStartNodes(startNodes, g1);

    std::set<std::shared_ptr<Node>> true_result;
    true_result.insert(FC1);
    true_result.insert(FC2);
    true_result.insert(FC3);

    // Test 2 : Return an empty set when the start nodes are wrong
    std::vector<std::shared_ptr<Node>> wrong_startNodes;
    std::set<std::shared_ptr<Node>> wrong_start_result;
    std::set<std::shared_ptr<Node>> empty_result;

    wrong_startNodes.push_back(Random0);
    wrong_start_result = GReg.matchFromStartNodes(wrong_startNodes, g1);

    // Perform tests
    REQUIRE(result == true_result);
    REQUIRE(wrong_start_result == empty_result);
}
*/

TEST_CASE("Function match | ALL matches of Nodes sequence", "[GRegex]") {
    // init all input for GRegex
    // Nodes Regex map : std::map<std::string,NodeRegex*>& nodesRegex
    // Sequential Regex vector : std::vector<std::string>& seqRegexps

    // init the Nodes Regex map 
    std::map<std::string,NodeRegex*> nodesRegex ;
    std::vector<std::string> nodeTypeKey {"GEMM"};
    for (const std::string& key : nodeTypeKey) {
        nodesRegex[key] = new NodeRegex(key); 
    }

    // init the Sequential Regex vector
    std::vector<std::string> seqRegex;
    seqRegex.push_back("GEMM;");

    // Instanciate a GRegex
    GRegex GReg(nodesRegex, seqRegex);

    //init the input graph
    std::shared_ptr<GraphView> graphToMatch = std::make_shared<GraphView>("TestGraph");
    std::shared_ptr<Node> Random0 = GenericOperator("Random", 1, 1, 1);
    std::shared_ptr<Node> GEMM1 = GenericOperator("GEMM", 1, 1, 1);
    std::shared_ptr<Node> ReLU1 = GenericOperator("ReLU", 1, 1, 1);
    std::shared_ptr<Node> GEMM2 = GenericOperator("GEMM", 1, 1, 1);
    std::shared_ptr<Node> GEMM3 = GenericOperator("GEMM", 1, 1, 1);
    std::shared_ptr<Node> ReLU2 = GenericOperator("ReLU", 1, 1, 1);
    std::shared_ptr<Node> Random = GenericOperator("Random", 1, 1, 1);

    graphToMatch->add(Random0);
    graphToMatch->addChild(GEMM1, Random0);
    graphToMatch->addChild(ReLU1, GEMM1);
    graphToMatch->addChild(GEMM2, ReLU1);
    graphToMatch->addChild(GEMM3, GEMM2);
    graphToMatch->addChild(ReLU2, GEMM3);
    graphToMatch->addChild(Random, ReLU2);

    
    //std::vector<std::pair<std::vector<NodeTmp>,std::set<NodeTmp>>> matchs = GReg.match(graphToMatch);
    //std::set<std::pair<std::vector<NodeTmp>,std::set<NodeTmp>>> matchs = GReg.match(graphToMatch);
    Match matches = GReg.match(graphToMatch);

    size_t nb = matches.getNbMatch();
    std::vector<std::vector<NodeTmp>> gm_startnodes = matches.getStartNodes();
    std::vector<std::set<NodeTmp>> gm_matchnodes = matches.getMatchNodes();

    std::set<std::pair<std::vector<NodeTmp>,std::set<NodeTmp>>> matchs;

    for (size_t i = 0; i < nb; ++i) {
        matchs.insert(std::make_pair(gm_startnodes[i], gm_matchnodes[i]));
    }

    //std::vector<std::pair<std::vector<NodeTmp>,std::set<NodeTmp>>> toMatchs ;
    std::set<std::pair<std::vector<NodeTmp>,std::set<NodeTmp>>> toMatchs ;
    // Carefull : as the assert is on a vector, the Order of match matters
    std::vector<NodeTmp> startNode = {GEMM1};
    std::set<NodeTmp> matchNode = {GEMM1};
    //toMatchs.push_back(std::make_pair(startNode,matchNode));
    toMatchs.insert(std::make_pair(startNode,matchNode));
    
    std::vector<NodeTmp> startNode2 = {GEMM2};
    std::set<NodeTmp> matchNode2 = {GEMM2};
    //toMatchs.push_back(std::make_pair(startNode2,matchNode2));
    toMatchs.insert(std::make_pair(startNode2,matchNode2));
    
    std::vector<NodeTmp> startNode3 = {GEMM3};
    std::set<NodeTmp> matchNode3 = {GEMM3};
    //toMatchs.push_back(std::make_pair(startNode3,matchNode3));
    toMatchs.insert(std::make_pair(startNode3,matchNode3));

    REQUIRE(matchs == toMatchs);
    REQUIRE(nb == 3);
}


