#include <cassert>
#include <map>
#include <memory>
#include <string>

#include <catch2/catch_test_macros.hpp>

#include "backend/OperatorImpl.hpp"
#include "data/Tensor.hpp"
#include "graph/GraphView.hpp"
#include "operator/Conv.hpp"
#include "operator/GenericOperator.hpp"
#include "operator/Producer.hpp"

using namespace Aidge;

TEST_CASE("[core/graph] GraphView(Constructor)") {
  std::shared_ptr<GraphView> g0 = std::make_shared<GraphView>();
  std::shared_ptr<GraphView> g1 = std::make_shared<GraphView>("G1");
  REQUIRE(g0 != nullptr);
  REQUIRE(g1 != nullptr);
}

TEST_CASE("[core/graph] GraphView(add)") {
  SECTION("Node alone") {
    std::shared_ptr<GraphView> g = std::make_shared<GraphView>("TestGraph");
    std::shared_ptr<Node> GOp1 = GenericOperator("Fictive", 0, 0, 0, "Gop1");
    g->add(GOp1);
    std::shared_ptr<Node> GOp2 = GenericOperator("Fictive", 0, 0, 1, "Gop2");
    g->add(GOp2);
    std::shared_ptr<Node> GOp3 = GenericOperator("Fictive", 1, 1, 0, "Gop3");
    g->add(GOp3);
    std::shared_ptr<Node> GOp4 = GenericOperator("Fictive", 0, 1, 0, "Gop4");
    g->add(GOp4);
    std::shared_ptr<Node> GOp5 = GenericOperator("Fictive", 1, 1, 1, "Gop5");
    g->add(GOp5);
    std::shared_ptr<Node> GOp6 = GenericOperator("Fictive", 1, 2, 1, "Gop6");
    g->add(GOp6);
  }

  SECTION("Several Nodes") {
    std::shared_ptr<GraphView> g = std::make_shared<GraphView>("TestGraph");
    // should automaticaly add parents for learnable parameters
    std::shared_ptr<Node> GOp1 = GenericOperator("Fictive", 0, 1, 1, "Gop1");
    std::shared_ptr<Node> GOp1parent = GenericOperator("Fictive", 0, 0, 1, "Gop1parent");
    GOp1parent->addChild(GOp1, 0, 0);
    g->add(GOp1);
    REQUIRE(g->getNodes() == std::set<std::shared_ptr<Node>>({GOp1, GOp1parent}));

    // there should be no deplicates
    g->add(GOp1);
    REQUIRE(g->getNodes() == std::set<std::shared_ptr<Node>>({GOp1, GOp1parent}));
  }

  SECTION("Initializer list ofr Node") {
    std::shared_ptr<GraphView> g = std::make_shared<GraphView>("TestGraph");
    std::shared_ptr<Node> GOp1 = GenericOperator("Fictive", 0, 0, 0, "Gop1");
    std::shared_ptr<Node> GOp2 = GenericOperator("Fictive", 0, 0, 0, "Gop2");
    g->add({GOp1, GOp1, GOp2});
    REQUIRE(g->getNodes() == std::set<std::shared_ptr<Node>>({GOp1, GOp2}));
  }

  SECTION("another GraphView") {
    std::shared_ptr<GraphView> g1 = std::make_shared<GraphView>("TestGraph-1");
    std::shared_ptr<GraphView> g2 = std::make_shared<GraphView>("TestGraph-2");
    auto conv = GenericOperator("Conv", 1, 1, 1, "c");
    auto conv1 = GenericOperator("Conv", 1, 1, 1, "c1");
    auto conv2 = GenericOperator("Conv", 1, 1, 1, "c2");
    auto conv3 = GenericOperator("Conv", 1, 1, 1, "c3");
    auto conv4 = GenericOperator("Conv", 1, 1, 1, "c4");
    conv->addChild(conv1);
    conv1->addChild(conv2);
    conv2->addChild(conv3);
    conv3->addChild(conv4);
    g1->add({conv, conv1, conv2, conv3, conv4});
    g2->add(g1);
    REQUIRE(((g1->getNodes() == g2->getNodes()) && (g2->getNodes() == std::set<std::shared_ptr<Node>>({conv, conv1, conv2, conv3, conv4}))));
    REQUIRE(((g1->inputNodes() == g2->inputNodes()) &&
           (g2->inputNodes() == std::set<std::shared_ptr<Node>>({conv}))));
    REQUIRE(((g1->outputNodes() == g2->outputNodes()) &&
           (g2->outputNodes() == std::set<std::shared_ptr<Node>>({conv4}))));
  }
}

TEST_CASE("[core/graph] GraphView(addChild)") {
  std::shared_ptr<GraphView> g1 = std::make_shared<GraphView>("TestGraph");
  std::shared_ptr<Node> conv = GenericOperator("Conv", 1, 1, 1, "c");
  std::shared_ptr<Node> conv1 = GenericOperator("Conv", 1, 1, 1, "c1");
  std::shared_ptr<Node> conv2 = GenericOperator("Conv", 1, 1, 1, "c2");
  std::shared_ptr<Node> conv3 = GenericOperator("Conv", 1, 1, 1, "c3");
  std::shared_ptr<Node> conv3_5 = GenericOperator("Conv", 1, 1, 1, "c3.5");
  std::shared_ptr<Node> conv4 = GenericOperator("Conv", 1, 1, 1, "c4");
  std::shared_ptr<Node> conv5 = GenericOperator("Conv", 1, 1, 1, "c5");

  g1->add(conv);
  SECTION("add(node)") {
    REQUIRE(g1->inputNodes() == std::set<std::shared_ptr<Node>>{conv});
    REQUIRE(g1->outputNodes() == std::set<std::shared_ptr<Node>>{conv});
  }
  g1->addChild(conv1, "c");
  SECTION("add(node, outputNodeName)") {
    REQUIRE(g1->inputNodes() == std::set<std::shared_ptr<Node>>{conv});
    REQUIRE(g1->outputNodes() == std::set<std::shared_ptr<Node>>{conv1});
    REQUIRE(conv->getChildren() == std::set<std::shared_ptr<Node>>({conv1}));
    REQUIRE(conv1->getParents() == std::vector<std::shared_ptr<Node>>({conv}));
  }
  g1->addChild(conv2, "c1", 0);
  SECTION("add(node, pair<outputNodeName, outID>)") {
    REQUIRE(g1->inputNodes() == std::set<std::shared_ptr<Node>>{conv});
    REQUIRE(g1->outputNodes() == std::set<std::shared_ptr<Node>>{conv2});
    REQUIRE(conv1->getChildren() == std::set<std::shared_ptr<Node>>({conv2}));
    REQUIRE(conv2->getParents() == std::vector<std::shared_ptr<Node>>({conv1}));
  }
  g1->addChild(conv3, "c2", 0, 0);
  SECTION("add(node, list(outputNodeName))") {
    REQUIRE(g1->inputNodes() == std::set<std::shared_ptr<Node>>{conv});
    REQUIRE(g1->outputNodes() == std::set<std::shared_ptr<Node>>{conv3});
    REQUIRE(conv2->getChildren() == std::set<std::shared_ptr<Node>>({conv3}));
    REQUIRE(conv3->getParents() == std::vector<std::shared_ptr<Node>>({conv2}));
  }
  g1->addChild(conv3_5, conv3);
  SECTION("add(node, list(outputNodeName))") {
    REQUIRE(g1->inputNodes() == std::set<std::shared_ptr<Node>>{conv});
    REQUIRE(g1->outputNodes() == std::set<std::shared_ptr<Node>>{conv3_5});
    REQUIRE(conv3->getChildren() == std::set<std::shared_ptr<Node>>({conv3_5}));
    REQUIRE(conv3_5->getParents() ==
            std::vector<std::shared_ptr<Node>>({conv3}));
  }
  g1->addChild(conv4, conv3_5, 0);
  SECTION("add(node, vector<pair<outputNodeName, outID>>)") {
    REQUIRE(g1->inputNodes() == std::set<std::shared_ptr<Node>>{conv});
    REQUIRE(g1->outputNodes() == std::set<std::shared_ptr<Node>>{conv4});
    REQUIRE(conv3_5->getChildren() == std::set<std::shared_ptr<Node>>({conv4}));
    REQUIRE(conv4->getParents() ==
            std::vector<std::shared_ptr<Node>>({conv3_5}));
  }
  g1->addChild(conv5, conv4, 0, 0);
  SECTION("add(node, vector<pair<outputNodeName, outID>>)") {
    REQUIRE(g1->inputNodes() == std::set<std::shared_ptr<Node>>{conv});
    REQUIRE(g1->outputNodes() == std::set<std::shared_ptr<Node>>{conv5});
    REQUIRE(conv4->getChildren() == std::set<std::shared_ptr<Node>>({conv5}));
    REQUIRE(conv5->getParents() == std::vector<std::shared_ptr<Node>>({conv4}));
  }
  std::set<std::shared_ptr<Node>> requiredNodes = {conv,    conv1, conv2, conv3,
                                                   conv3_5, conv4, conv5};
  REQUIRE(g1->getNodes() == requiredNodes);
  REQUIRE(g1->getChildren(conv3) == std::set<std::shared_ptr<Node>>({conv3_5}));
}

TEST_CASE("[core/graph] GraphView(inputs)") {
  auto g1 = std::make_shared<GraphView>("TestGraph");
  std::shared_ptr<Node> conv = Conv(3, 32, {3, 3});
  g1->add(conv);

  REQUIRE(g1->inputs() == conv->inputs());
}

TEST_CASE("[core/graph] GraphView(outputs)") {
  std::shared_ptr<GraphView> g1 = std::make_shared<GraphView>("TestGraph");
  std::shared_ptr<Node> conv = Conv(3, 32, {3, 3});
  g1->add(conv);

  REQUIRE(g1->outputs() == conv->outputs());
}

TEST_CASE("[core/graph] GraphView(save)") {
  std::shared_ptr<GraphView> g1 = std::make_shared<GraphView>("TestGraph");
  std::shared_ptr<Node> conv = GenericOperator("Conv", 1, 1, 1, "c");
  std::shared_ptr<Node> conv1 = GenericOperator("Conv", 1, 1, 1, "c1");
  std::shared_ptr<Node> conv2 = GenericOperator("Conv", 1, 1, 1, "c2");
  std::shared_ptr<Node> conv3 = GenericOperator("Conv", 1, 1, 1, "c3");
  std::shared_ptr<Node> conv4 = GenericOperator("Conv", 1, 1, 1, "c4");
  std::shared_ptr<Node> conv5 = GenericOperator("Conv", 1, 1, 1, "c5");

  g1->add(conv);
  g1->addChild(conv1, "c");
  g1->addChild(conv2, "c1", 0);
  g1->addChild(conv3, "c2");
  g1->addChild(conv4, "c3", 0);
  g1->addChild(conv5, "c4", 0, 0);

  g1->save("./graphExample");
  printf("File saved in ./graphExample.md\n");
}

TEST_CASE("[core/graph] GraphView(resetConnections)") {
  SECTION("disconnect data iput") {
    std::shared_ptr<Node> conv = GenericOperator("Conv", 1, 1, 1, "c");
    std::shared_ptr<Node> conv1 = GenericOperator("Conv", 1, 3, 1, "c1");
    std::shared_ptr<Node> conv2 = GenericOperator("Conv", 1, 1, 1, "c2");
    std::shared_ptr<Node> prod1 = GenericOperator("Prod", 0, 0, 1, "p1");
    std::shared_ptr<Node> prod2 = GenericOperator("Prod", 0, 0, 1, "p2");
    conv->addChild(conv1);
    prod1->addChild(conv1,0,1);
    prod2->addChild(conv1,0,2);
    conv1->addChild(conv2);

    conv1->resetConnections(false);

    REQUIRE(conv->output(0).size() == 0);
    for (std::size_t i = 0; i < conv1->nbDataInputs(); ++i) {
      REQUIRE((conv1->input(i) == std::pair<std::shared_ptr<Node>, IOIndex_t>(nullptr, gk_IODefaultIndex)));
    }
    REQUIRE((conv1->input(1) == std::pair<std::shared_ptr<Node>, IOIndex_t>(prod1, 0)));
    REQUIRE((conv1->input(2) == std::pair<std::shared_ptr<Node>, IOIndex_t>(prod2, 0)));
    REQUIRE((conv2->input(0) == std::pair<std::shared_ptr<Node>, IOIndex_t>(nullptr, gk_IODefaultIndex)));
    for (std::size_t i = 0; i < conv1->nbOutputs(); ++i) {
      REQUIRE(conv->output(i).size() == 0U);
    }
  }

  SECTION("disconnect data iput + learnable parameters") {
    std::shared_ptr<Node> conv = GenericOperator("Conv", 1, 1, 1, "c");
    std::shared_ptr<Node> conv1 = GenericOperator("Conv", 1, 3, 1, "c1");
    std::shared_ptr<Node> conv2 = GenericOperator("Conv", 1, 1, 1, "c2");
    std::shared_ptr<Node> prod1 = GenericOperator("Prod", 0, 0, 1, "p1");
    std::shared_ptr<Node> prod2 = GenericOperator("Prod", 0, 0, 1, "p2");
    conv->addChild(conv1);
    prod1->addChild(conv1,0,1);
    prod2->addChild(conv1,0,2);
    conv1->addChild(conv2);

    conv1->resetConnections(true);

    REQUIRE(conv->output(0).size() == 0);
    for (std::size_t i = 0; i < conv1->nbInputs(); ++i) {
      REQUIRE((conv1->input(i) == std::pair<std::shared_ptr<Node>, IOIndex_t>(nullptr, gk_IODefaultIndex)));
    }
    REQUIRE((conv2->input(0) == std::pair<std::shared_ptr<Node>, IOIndex_t>(nullptr, gk_IODefaultIndex)));
    for (std::size_t i = 0; i < conv1->nbOutputs(); ++i) {
      REQUIRE(conv->output(i).size() == 0U);
    }
  }
}

TEST_CASE("Graph Forward dims", "[GraphView]") {
  auto dataProvider = Producer({16, 3, 224, 224}, "dataProvider");
  auto conv1 = Conv(3, 32, {3, 3}, "conv1");
  auto conv2 = Conv(32, 64, {3, 3}, "conv2");
  auto conv3 = Conv(64, 10, {1, 1}, "conv3");
  auto g = std::make_shared<GraphView>("TestGraph");
  dataProvider->addChild(conv1, 0);
  g->add(conv1);
  g->addChild(conv2, conv1, 0);
  g->addChild(conv3, conv2, 0);
  g->save("graphForwardDims");
  g->forwardDims();

  SECTION("Check input-output connections") {
    REQUIRE(dataProvider->getOperator()->getOutput(0) ==
            conv1->getOperator()->getInput(0));
    REQUIRE(conv1->getOperator()->getInput(1) ==
            g->getNode("conv1_w")->getOperator()->getOutput(0));
    REQUIRE(conv1->getOperator()->getInput(2) ==
            g->getNode("conv1_b")->getOperator()->getOutput(0));
    REQUIRE(conv1->getOperator()->getOutput(0) ==
            conv2->getOperator()->getInput(0));
    REQUIRE(conv2->getOperator()->getInput(1) ==
            g->getNode("conv2_w")->getOperator()->getOutput(0));
    REQUIRE(conv2->getOperator()->getInput(2) ==
            g->getNode("conv2_b")->getOperator()->getOutput(0));
    REQUIRE(conv2->getOperator()->getOutput(0) ==
            conv3->getOperator()->getInput(0));
    REQUIRE(conv3->getOperator()->getInput(1) ==
            g->getNode("conv3_w")->getOperator()->getOutput(0));
    REQUIRE(conv3->getOperator()->getInput(2) ==
            g->getNode("conv3_b")->getOperator()->getOutput(0));
  }

  SECTION("Check forwarded dims") {
    REQUIRE(std::static_pointer_cast<Tensor>(conv1->getOperator()->getOutput(0))
                ->dims() == std::vector<DimSize_t>({16, 32, 222, 222}));
    REQUIRE(std::static_pointer_cast<Tensor>(conv2->getOperator()->getOutput(0))
                ->dims() == std::vector<DimSize_t>({16, 64, 220, 220}));
  }
}

TEST_CASE("[core/graph] GraphView(replaceWith)") {
  // create original graph
  std::shared_ptr<GraphView> g = std::make_shared<GraphView>("TestGraph");
  auto otherInput = GenericOperator("Producer", 0, 0, 1, "other_input");
  auto matmulWeight = GenericOperator("Producer", 0, 0, 1, "matmul_w");
  auto addBias = GenericOperator("Producer", 0, 0, 1, "add_b");
  auto other1 = GenericOperator("Other", 1, 1, 1, "other1");
  auto other2 = GenericOperator("Other", 1, 1, 1, "other2");
  auto matmul = GenericOperator("MatMul", 1, 2, 1, "matmul");
  auto add = GenericOperator("Add", 1, 2, 1, "add");
  otherInput->addChild(other1);
  other1->addChild(matmul);
  matmul->addChild(add);
  add->addChild(other2);
  matmulWeight->addChild(matmul, 0, 1);
  addBias->addChild(add, 0, 1);
  g->add({other1, matmul, add, other2});
  REQUIRE(g->getNodes() == std::set<std::shared_ptr<Node>>({matmulWeight, addBias, other1, other2, matmul, add}));

  // create graph to replace
  std::shared_ptr<GraphView> nodeToReplace = std::make_shared<GraphView>();
  nodeToReplace->add({matmul, add}, false);

  // create replacing graph
  std::shared_ptr<Node> newNode = GenericOperator("FC", 1, 3, 1, "fc");
  other1->addChild(newNode);
  matmulWeight->addChild(newNode, 0, 1);
  addBias->addChild(newNode, 0, 2);

  // replace
  nodeToReplace->replaceWith({newNode});

  REQUIRE(g->getNodes() == std::set<std::shared_ptr<Node>>({matmulWeight, addBias, other1, other2, newNode}));
}